@@ -1,5 +1,5 @@
 /*
-* Copyright (c) 2022-2023, The PurpleI2P Project
+* Copyright (c) 2022-2024, The PurpleI2P Project
 *
 * This file is part of Purple i2pd project and licensed under BSD3
 *
@@ -25,7 +25,7 @@ namespace transport
 		m_TerminationTimer (GetService ()), m_CleanupTimer (GetService ()), m_ResendTimer (GetService ()),
 		m_IntroducersUpdateTimer (GetService ()), m_IntroducersUpdateTimerV6 (GetService ()),
 		m_IsPublished (true), m_IsSyncClockFromPeers (true), m_PendingTimeOffset (0),
-		m_IsThroughProxy (false)
+		m_Rng(i2p::util::GetMonotonicMicroseconds ()%1000000LL), m_IsThroughProxy (false)
 	{
 	}
 
@@ -152,8 +152,11 @@ namespace transport
 		m_SessionsByRouterHash.clear ();
 		m_PendingOutgoingSessions.clear ();
 		m_Relays.clear ();
+		m_PeerTests.clear ();
 		m_Introducers.clear ();
 		m_IntroducersV6.clear ();
+		m_ConnectedRecently.clear ();
+		m_RequestedPeerTests.clear ();
 	}
 
 	void SSU2Server::SetLocalAddress (const boost::asio::ip::address& localAddress)
@@ -210,27 +213,63 @@ namespace transport
 		return ep.port ();
 	}
 
-	void SSU2Server::AdjustTimeOffset (int64_t offset)
+	bool SSU2Server::IsConnectedRecently (const boost::asio::ip::udp::endpoint& ep)
+	{
+		if (!ep.port () || ep.address ().is_unspecified ()) return false;
+		auto it = m_ConnectedRecently.find (ep);
+		if (it != m_ConnectedRecently.end ())
+		{	
+			if (i2p::util::GetSecondsSinceEpoch () <= it->second + SSU2_HOLE_PUNCH_EXPIRATION)
+				return true;
+			else
+				m_ConnectedRecently.erase (it);
+		}	
+		return false;
+	}	
+
+	void SSU2Server::AddConnectedRecently (const boost::asio::ip::udp::endpoint& ep, uint64_t ts)
+	{
+		if (!ep.port () || ep.address ().is_unspecified () || 
+		    i2p::util::GetSecondsSinceEpoch () > ts + SSU2_HOLE_PUNCH_EXPIRATION) return;
+		auto [it, added] = m_ConnectedRecently.try_emplace (ep, ts);
+		if (!added && ts > it->second)
+			it->second = ts; // renew timestamp of existing endpoint
+	}	
+		
+	void SSU2Server::AdjustTimeOffset (int64_t offset, std::shared_ptr<const i2p::data::IdentityEx> from)
 	{
 		if (offset)
 		{	
 			if (m_PendingTimeOffset) // one more
 			{	
-				if (std::abs (m_PendingTimeOffset - offset) < SSU2_CLOCK_SKEW)
+				if (m_PendingTimeOffsetFrom && from && 
+					m_PendingTimeOffsetFrom->GetIdentHash ().GetLL()[0] != from->GetIdentHash ().GetLL()[0]) // from different routers
 				{	
-					offset = (m_PendingTimeOffset + offset)/2; // average
-					LogPrint (eLogWarning, "SSU2: Clock adjusted by ", offset, " seconds");
-					i2p::util::AdjustTimeOffset (offset);
-				}	
+					if (std::abs (m_PendingTimeOffset - offset) < SSU2_CLOCK_SKEW)
+					{	
+						offset = (m_PendingTimeOffset + offset)/2; // average
+						LogPrint (eLogWarning, "SSU2: Clock adjusted by ", offset, " seconds");
+						i2p::util::AdjustTimeOffset (offset);
+					}	
+					else
+						LogPrint (eLogWarning, "SSU2: Time offsets are too different. Clock not adjusted");
+					m_PendingTimeOffset = 0;
+					m_PendingTimeOffsetFrom = nullptr;
+				}
 				else
-					LogPrint (eLogWarning, "SSU2: Time offsets are too different. Clock not adjusted");
-				m_PendingTimeOffset = 0;
+					LogPrint (eLogWarning, "SSU2: Time offsets from same router. Clock not adjusted");
 			}
 			else
+			{	
 				m_PendingTimeOffset = offset; // first 
+				m_PendingTimeOffsetFrom = from;
+			}	
 		}	
 		else
+		{	
 			m_PendingTimeOffset = 0; // reset
+			m_PendingTimeOffsetFrom = nullptr;
+		}	
 	}	
 		
 	boost::asio::ip::udp::socket& SSU2Server::OpenSocket (const boost::asio::ip::udp::endpoint& localEndpoint)
@@ -243,15 +282,49 @@ namespace transport
 			socket.open (localEndpoint.protocol ());
 			if (localEndpoint.address ().is_v6 ())
 				socket.set_option (boost::asio::ip::v6_only (true));
-			socket.set_option (boost::asio::socket_base::receive_buffer_size (SSU2_SOCKET_RECEIVE_BUFFER_SIZE));
-			socket.set_option (boost::asio::socket_base::send_buffer_size (SSU2_SOCKET_SEND_BUFFER_SIZE));
+
+			uint64_t bufferSize = i2p::context.GetBandwidthLimit() * 1024 / 5; // max lag = 200ms
+			bufferSize = std::max(SSU2_SOCKET_MIN_BUFFER_SIZE, std::min(bufferSize, SSU2_SOCKET_MAX_BUFFER_SIZE));
+
+			boost::asio::socket_base::receive_buffer_size receiveBufferSizeSet (bufferSize);
+			boost::asio::socket_base::send_buffer_size sendBufferSizeSet (bufferSize);
+			socket.set_option (receiveBufferSizeSet);
+			socket.set_option (sendBufferSizeSet);
+			boost::asio::socket_base::receive_buffer_size receiveBufferSizeGet;
+			boost::asio::socket_base::send_buffer_size sendBufferSizeGet;
+			socket.get_option (receiveBufferSizeGet);
+			socket.get_option (sendBufferSizeGet);
+			if (receiveBufferSizeGet.value () != receiveBufferSizeSet.value () ||
+				sendBufferSizeGet.value () != sendBufferSizeSet.value ())
+			{
+				LogPrint (eLogWarning, "SSU2: Socket receive buffer size: requested = ",
+					receiveBufferSizeSet.value (), ", got = ", receiveBufferSizeGet.value ());
+				LogPrint (eLogWarning, "SSU2: Socket send buffer size: requested = ",
+					sendBufferSizeSet.value (), ", got = ", sendBufferSizeGet.value ());
+			}
+			else
+			{
+				LogPrint (eLogInfo, "SSU2: Socket receive buffer size: ", receiveBufferSizeGet.value ());
+				LogPrint (eLogInfo, "SSU2: Socket send buffer size: ", sendBufferSizeGet.value ());
+			}
+
+			socket.non_blocking (true);
+		}
+		catch (std::exception& ex )
+		{
+			LogPrint (eLogCritical, "SSU2: Failed to open socket on ", localEndpoint.address (), ": ", ex.what());
+			ThrowFatal ("Unable to start SSU2 transport on ", localEndpoint.address (), ": ", ex.what ());
+			return socket;
+		}
+		try
+		{
 			socket.bind (localEndpoint);
 			LogPrint (eLogInfo, "SSU2: Start listening on ", localEndpoint);
 		}
 		catch (std::exception& ex )
 		{
-			LogPrint (eLogCritical, "SSU2: Failed to bind to ", localEndpoint, ": ", ex.what());
-			ThrowFatal ("Unable to start SSU2 transport on ", localEndpoint, ": ", ex.what ());
+			LogPrint (eLogWarning, "SSU2: Failed to bind to ", localEndpoint, ": ", ex.what(), ". Actual endpoint is ", socket.local_endpoint ());
+			// we can continue without binding being firewalled
 		}
 		return socket;
 	}
@@ -283,22 +356,38 @@ namespace transport
 		// but better to find out which host were sent it and mark that router as unreachable
 		{
 			i2p::transport::transports.UpdateReceivedBytes (bytes_transferred);
+			if (bytes_transferred < SSU2_MIN_RECEIVED_PACKET_SIZE)
+			{
+				// drop too short packets
+				m_PacketsPool.ReleaseMt (packet);
+				Receive (socket);
+				return;
+			}	
 			packet->len = bytes_transferred;
-
+			
 			boost::system::error_code ec;
 			size_t moreBytes = socket.available (ec);
 			if (!ec && moreBytes)
 			{
-				std::vector<Packet *> packets;
-				packets.push_back (packet);
-				while (moreBytes && packets.size () < 32)
+				auto packets = m_PacketsArrayPool.AcquireMt ();
+				packets->AddPacket (packet);
+				while (moreBytes && packets->numPackets < SSU2_MAX_NUM_PACKETS_PER_BATCH)
 				{
 					packet = m_PacketsPool.AcquireMt ();
 					packet->len = socket.receive_from (boost::asio::buffer (packet->buf, SSU2_MAX_PACKET_SIZE), packet->from, 0, ec);
 					if (!ec)
 					{
 						i2p::transport::transports.UpdateReceivedBytes (packet->len);
-						packets.push_back (packet);
+						if (packet->len >= SSU2_MIN_RECEIVED_PACKET_SIZE)
+						{	
+							if (!packets->AddPacket (packet))
+							{
+								LogPrint (eLogError, "SSU2: Received packets array is full");
+								m_PacketsPool.ReleaseMt (packet);
+							}	
+						}	
+						else // drop too short packets
+							m_PacketsPool.ReleaseMt (packet);
 						moreBytes = socket.available(ec);
 						if (ec) break;
 					}
@@ -353,15 +442,23 @@ namespace transport
 		}
 	}
 
-	void SSU2Server::HandleReceivedPackets (std::vector<Packet *> packets)
+	void SSU2Server::HandleReceivedPackets (Packets * packets)
 	{
+		if (!packets) return;
 		if (m_IsThroughProxy)
-			for (auto& packet: packets)
+			for (size_t i = 0; i < packets->numPackets; i++)
+			{
+				auto& packet = (*packets)[i];
 				ProcessNextPacketFromProxy (packet->buf, packet->len);
+			}	
 		else
-			for (auto& packet: packets)
+			for (size_t i = 0; i < packets->numPackets; i++)
+			{
+				auto& packet = (*packets)[i];
 				ProcessNextPacket (packet->buf, packet->len, packet->from);
-		m_PacketsPool.ReleaseMt (packets);
+			}	
+		m_PacketsPool.ReleaseMt (packets->data (), packets->numPackets);
+		m_PacketsArrayPool.ReleaseMt (packets);
 		if (m_LastSession && m_LastSession->GetState () != eSSU2SessionStateTerminated)
 			m_LastSession->FlushData ();
 	}
@@ -371,7 +468,8 @@ namespace transport
 		if (session)
 		{
 			m_Sessions.emplace (session->GetConnID (), session);
-			AddSessionByRouterHash (session);
+			if (session->GetState () != eSSU2SessionStatePeerTest)
+				AddSessionByRouterHash (session);
 		}
 	}
 
@@ -380,15 +478,28 @@ namespace transport
 		auto it = m_Sessions.find (connID);
 		if (it != m_Sessions.end ())
 		{
-			auto ident = it->second->GetRemoteIdentity ();
-			if (ident)
-				m_SessionsByRouterHash.erase (ident->GetIdentHash ());
+			if (it->second->GetState () != eSSU2SessionStatePeerTest)
+			{	
+				auto ident = it->second->GetRemoteIdentity ();
+				if (ident)
+				{
+					std::lock_guard<std::mutex> l(m_SessionsByRouterHashMutex);
+					auto it1 = m_SessionsByRouterHash.find (ident->GetIdentHash ());
+					if (it1 != m_SessionsByRouterHash.end () && it->second == it1->second.lock ())
+						m_SessionsByRouterHash.erase (it1);
+				}	
+			}
 			if (m_LastSession == it->second)
 				m_LastSession = nullptr;
 			m_Sessions.erase (it);
 		}
 	}
 
+	void SSU2Server::RequestRemoveSession (uint64_t connID)
+	{
+		GetService ().post ([connID, this]() { RemoveSession (connID); });
+	}	
+		
 	void SSU2Server::AddSessionByRouterHash (std::shared_ptr<SSU2Session> session)
 	{
 		if (session)
@@ -396,16 +507,26 @@ namespace transport
 			auto ident = session->GetRemoteIdentity ();
 			if (ident)
 			{
-				auto ret = m_SessionsByRouterHash.emplace (ident->GetIdentHash (), session);
-				if (!ret.second)
+				std::shared_ptr<SSU2Session> oldSession;
 				{
+					std::lock_guard<std::mutex> l(m_SessionsByRouterHashMutex);
+					auto ret = m_SessionsByRouterHash.emplace (ident->GetIdentHash (), session);
+					if (!ret.second)
+					{
+						oldSession = ret.first->second.lock ();
+						// update session
+						ret.first->second = session;
+					}	
+				}	
+				if (oldSession && oldSession != session)
+				{	
 					// session already exists
 					LogPrint (eLogWarning, "SSU2: Session to ", ident->GetIdentHash ().ToBase64 (), " already exists");
+					// move unsent msgs to new session
+					oldSession->MoveSendQueue (session);
 					// terminate existing
-					GetService ().post (std::bind (&SSU2Session::RequestTermination, ret.first->second, eSSU2TerminationReasonReplacedByNewSession));
-					// update session
-					ret.first->second = session;
-				}
+					GetService ().post (std::bind (&SSU2Session::RequestTermination, oldSession, eSSU2TerminationReasonReplacedByNewSession));
+				}	
 			}
 		}
 	}
@@ -413,21 +534,30 @@ namespace transport
 	bool SSU2Server::AddPendingOutgoingSession (std::shared_ptr<SSU2Session> session)
 	{
 		if (!session) return false;
-		std::unique_lock<std::mutex> l(m_PendingOutgoingSessionsMutex);
+		std::lock_guard<std::mutex> l(m_PendingOutgoingSessionsMutex);
 		return m_PendingOutgoingSessions.emplace (session->GetRemoteEndpoint (), session).second;
 	}
 
-	std::shared_ptr<SSU2Session> SSU2Server::FindSession (const i2p::data::IdentHash& ident) const
+	std::shared_ptr<SSU2Session> SSU2Server::FindSession (const i2p::data::IdentHash& ident)
 	{
+		std::lock_guard<std::mutex> l(m_SessionsByRouterHashMutex);
 		auto it = m_SessionsByRouterHash.find (ident);
 		if (it != m_SessionsByRouterHash.end ())
-			return it->second;
+		{	
+			if (!it->second.expired ())
+			{
+				auto s = it->second.lock ();
+				if (s && s->GetState () != eSSU2SessionStateTerminated)
+					return s;
+			}
+			m_SessionsByRouterHash.erase (it);
+		}	
 		return nullptr;
 	}
 
 	std::shared_ptr<SSU2Session> SSU2Server::FindPendingOutgoingSession (const boost::asio::ip::udp::endpoint& ep) const
 	{
-		std::unique_lock<std::mutex> l(m_PendingOutgoingSessionsMutex);
+		std::lock_guard<std::mutex> l(m_PendingOutgoingSessionsMutex);
 		auto it = m_PendingOutgoingSessions.find (ep);
 		if (it != m_PendingOutgoingSessions.end ())
 			return it->second;
@@ -436,22 +566,20 @@ namespace transport
 
 	void SSU2Server::RemovePendingOutgoingSession (const boost::asio::ip::udp::endpoint& ep)
 	{
-		std::unique_lock<std::mutex> l(m_PendingOutgoingSessionsMutex);
+		std::lock_guard<std::mutex> l(m_PendingOutgoingSessionsMutex);
 		m_PendingOutgoingSessions.erase (ep);
 	}
 
-	std::shared_ptr<SSU2Session> SSU2Server::GetRandomSession (
-		i2p::data::RouterInfo::CompatibleTransports remoteTransports, const i2p::data::IdentHash& excluded) const
+	std::shared_ptr<SSU2Session> SSU2Server::GetRandomPeerTestSession (
+		i2p::data::RouterInfo::CompatibleTransports remoteTransports, const i2p::data::IdentHash& excluded)
 	{
 		if (m_Sessions.empty ()) return nullptr;
-		uint16_t ind;
-		RAND_bytes ((uint8_t *)&ind, sizeof (ind));
-		ind %= m_Sessions.size ();
+		int ind = m_Rng () % m_Sessions.size ();
 		auto it = m_Sessions.begin ();
 		std::advance (it, ind);
 		while (it != m_Sessions.end ())
 		{
-			if ((it->second->GetRemoteTransports () & remoteTransports) &&
+			if (it->second->IsEstablished () && (it->second->GetRemotePeerTestTransports () & remoteTransports) &&
 			    it->second->GetRemoteIdentity ()->GetIdentHash () != excluded)
 				return it->second;
 			it++;
@@ -460,7 +588,7 @@ namespace transport
 		it = m_Sessions.begin ();
 		while (it != m_Sessions.end () && ind)
 		{
-			if ((it->second->GetRemoteTransports () & remoteTransports) &&
+			if (it->second->IsEstablished () && (it->second->GetRemotePeerTestTransports () & remoteTransports) &&
 			    it->second->GetRemoteIdentity ()->GetIdentHash () != excluded)
 				return it->second;
 			it++; ind--;
@@ -483,14 +611,51 @@ namespace transport
 		auto it = m_Relays.find (tag);
 		if (it != m_Relays.end ())
 		{
-			if (it->second->IsEstablished ())
-				return it->second;
-			else
-				m_Relays.erase (it);
+			if (!it->second.expired ())
+			{	
+				auto s = it->second.lock ();
+				if (s && s->IsEstablished ())
+					return s;
+			}	
+			m_Relays.erase (it);
 		}
 		return nullptr;
 	}
 
+	bool SSU2Server::AddPeerTest (uint32_t nonce, std::shared_ptr<SSU2Session> aliceSession, uint64_t ts)
+	{
+		return m_PeerTests.emplace (nonce, std::pair{ aliceSession, ts }).second;
+	}	
+
+	std::shared_ptr<SSU2Session> SSU2Server::GetPeerTest (uint32_t nonce)
+	{
+		auto it = m_PeerTests.find (nonce);
+		if (it != m_PeerTests.end ())
+		{
+			auto s = it->second.first.lock ();
+			m_PeerTests.erase (it);
+			return s;
+		}
+		return nullptr;
+	}	
+		
+	bool SSU2Server::AddRequestedPeerTest (uint32_t nonce, std::shared_ptr<SSU2PeerTestSession> session, uint64_t ts)
+	{
+		return m_RequestedPeerTests.emplace (nonce, std::pair{ session, ts }).second;
+	}
+		
+	std::shared_ptr<SSU2PeerTestSession> SSU2Server::GetRequestedPeerTest (uint32_t nonce)
+	{
+		auto it = m_RequestedPeerTests.find (nonce);
+		if (it != m_RequestedPeerTests.end ())
+		{
+			auto s = it->second.first.lock ();
+			m_RequestedPeerTests.erase (it);
+			return s;
+		}
+		return nullptr;
+	}	
+		
 	void SSU2Server::ProcessNextPacket (uint8_t * buf, size_t len, const boost::asio::ip::udp::endpoint& senderEndpoint)
 	{
 		if (len < 24) return;
@@ -560,13 +725,13 @@ namespace transport
 				if (it1->second->GetState () == eSSU2SessionStateSessionRequestSent &&
 					it1->second->ProcessSessionCreated (buf, len))
 				{
-					std::unique_lock<std::mutex> l(m_PendingOutgoingSessionsMutex);
+					std::lock_guard<std::mutex> l(m_PendingOutgoingSessionsMutex);
 					m_PendingOutgoingSessions.erase (it1); // we are done with that endpoint
 				}
 				else
 					it1->second->ProcessRetry (buf, len);
 			}
-			else if (!i2p::util::net::IsInReservedRange(senderEndpoint.address ()) && senderEndpoint.port ())
+			else if (!i2p::transport::transports.IsInReservedRange(senderEndpoint.address ()) && senderEndpoint.port ())
 			{
 				// assume new incoming session
 				auto session = std::make_shared<SSU2Session> (*this);
@@ -608,7 +773,10 @@ namespace transport
 		if (!ec)
 			i2p::transport::transports.UpdateSentBytes (headerLen + payloadLen);
 		else
-			LogPrint (eLogError, "SSU2: Send exception: ", ec.message (), " to ", to);
+		{
+			LogPrint (ec == boost::asio::error::would_block ? eLogInfo : eLogError,
+				"SSU2: Send exception: ", ec.message (), " to ", to);
+		}
 	}
 
 	void SSU2Server::Send (const uint8_t * header, size_t headerLen, const uint8_t * headerX, size_t headerXLen,
@@ -642,7 +810,10 @@ namespace transport
 		if (!ec)
 			i2p::transport::transports.UpdateSentBytes (headerLen + headerXLen + payloadLen);
 		else
-			LogPrint (eLogError, "SSU2: Send exception: ", ec.message (), " to ", to);
+		{
+			LogPrint (ec == boost::asio::error::would_block ? eLogInfo : eLogError,
+				"SSU2: Send exception: ", ec.message (), " to ", to);
+		}
 	}
 
 	bool SSU2Server::CreateSession (std::shared_ptr<const i2p::data::RouterInfo> router,
@@ -651,22 +822,19 @@ namespace transport
 		if (router && address)
 		{
 			// check if no session
-			auto it = m_SessionsByRouterHash.find (router->GetIdentHash ());
-			if (it != m_SessionsByRouterHash.end ())
+			auto existingSession = FindSession (router->GetIdentHash ());
+			if (existingSession)
 			{
 				// session with router found, trying to send peer test if requested
-				if (peerTest && it->second->IsEstablished ())
-				{
-					auto session = it->second;
-					GetService ().post ([session]() { session->SendPeerTest (); });
-				}
+				if (peerTest && existingSession->IsEstablished ())
+					GetService ().post ([existingSession]() { existingSession->SendPeerTest (); });
 				return false;
 			}
 			// check is no pending session
 			bool isValidEndpoint = !address->host.is_unspecified () && address->port;
 			if (isValidEndpoint)
 			{
-				if (i2p::util::net::IsInReservedRange(address->host)) return false;
+				if (i2p::transport::transports.IsInReservedRange(address->host)) return false;
 				auto s = FindPendingOutgoingSession (boost::asio::ip::udp::endpoint (address->host, address->port));
 				if (s)
 				{
@@ -709,82 +877,119 @@ namespace transport
 		auto address = session->GetAddress ();
 		if (!address) return;
 		session->WaitForIntroduction ();
+		auto ts = i2p::util::GetSecondsSinceEpoch ();
+		std::vector<int> indices; int i = 0;
 		// try to find existing session first
 		for (auto& it: address->ssu->introducers)
 		{
-			auto it1 = m_SessionsByRouterHash.find (it.iH);
-			if (it1 != m_SessionsByRouterHash.end ())
+			if (it.iTag && ts < it.iExp)
 			{
-				it1->second->Introduce (session, it.iTag);
-				return;
-			}
-		}
-		// we have to start a new session to an introducer
-		auto ts = i2p::util::GetSecondsSinceEpoch ();
-		std::shared_ptr<i2p::data::RouterInfo> r;
-		uint32_t relayTag = 0;
-		if (!address->ssu->introducers.empty ())
-		{
-			std::vector<int> indices;
-			for (int i = 0; i < (int)address->ssu->introducers.size (); i++) indices.push_back(i);
-			if (indices.size () > 1)
-				std::shuffle (indices.begin(), indices.end(), std::mt19937(std::random_device()()));
-
-			for (auto i: indices)
-			{
-				const auto& introducer = address->ssu->introducers[indices[i]];
-				if (introducer.iTag && ts < introducer.iExp)
+				auto s = FindSession (it.iH);
+				if (s)
 				{
-					r = i2p::data::netdb.FindRouter (introducer.iH);
-					if (r && r->IsReachableFrom (i2p::context.GetRouterInfo ()))
-					{
-						relayTag = introducer.iTag;
-						if (relayTag) break;
+					auto addr = s->GetAddress ();
+					if (addr && addr->IsIntroducer ())
+					{	
+						s->Introduce (session, it.iTag);
+						return;
 					}
 				}
+				else
+					indices.push_back(i);
+			}	
+			i++;
+		}
+		// we have to start a new session to an introducer
+		std::vector<i2p::data::IdentHash> newRouters;
+		std::shared_ptr<i2p::data::RouterInfo> r;
+		std::shared_ptr<const i2p::data::RouterInfo::Address> addr;
+		uint32_t relayTag = 0;
+		if (!indices.empty ())
+		{
+			if (indices.size () > 1)
+				std::shuffle (indices.begin(), indices.end(), m_Rng);
+
+			for (auto ind: indices)
+			{
+				const auto& introducer = address->ssu->introducers[ind];
+				// introducer is not expired, because in indices
+				r = i2p::data::netdb.FindRouter (introducer.iH);
+				if (r)
+				{	
+					if (r->IsPublishedOn (i2p::context.GetRouterInfo ().GetCompatibleTransports (false) & // outgoing
+					    (i2p::data::RouterInfo::eSSU2V4 | i2p::data::RouterInfo::eSSU2V6)))                  
+					{
+						relayTag = introducer.iTag;
+						addr = address->IsV6 () ? r->GetSSU2V6Address () : r->GetSSU2V4Address ();
+						if (addr && addr->IsIntroducer () && !addr->host.is_unspecified () && addr->port &&
+							!i2p::transport::transports.IsInReservedRange(addr->host))
+							break;
+						else
+						{
+							// address is invalid or not intrudcer, try another SSU2 address if exists
+							if (address->IsV4 ())
+							{
+								if (i2p::context.SupportsV6 ())
+									addr = r->GetSSU2V6Address ();
+							}	
+							else
+							{
+								if (i2p::context.SupportsV4 ())
+									addr = r->GetSSU2V4Address ();
+							}	
+							if (addr && addr->IsIntroducer () && !addr->host.is_unspecified () && addr->port &&
+								!i2p::transport::transports.IsInReservedRange(addr->host))
+								break;
+							else
+							{	
+								// all addresses are invalid, try next introducer
+								relayTag = 0;
+								addr = nullptr;
+								r = nullptr;
+							}	
+						}	
+					}
+					else
+						r = nullptr;
+				}	
+				else if (!i2p::data::IsRouterBanned (introducer.iH))
+					newRouters.push_back (introducer.iH);
 			}
 		}
 		if (r)
 		{
-			if (relayTag)
+			if (relayTag && addr)
 			{
 				// introducer and tag found connect to it through SSU2
-				auto addr = address->IsV6 () ? r->GetSSU2V6Address () : r->GetSSU2V4Address ();
-				if (addr)
+				auto s = FindPendingOutgoingSession (boost::asio::ip::udp::endpoint (addr->host, addr->port));
+				if (!s)
 				{
-					bool isValidEndpoint = !addr->host.is_unspecified () && addr->port &&
-						!i2p::util::net::IsInReservedRange(addr->host);
-					if (isValidEndpoint)
-					{
-						auto s = FindPendingOutgoingSession (boost::asio::ip::udp::endpoint (addr->host, addr->port));
-						if (!s)
-						{
-							s = std::make_shared<SSU2Session> (*this, r, addr);
-							s->SetOnEstablished ([session, s, relayTag]() { s->Introduce (session, relayTag); });
-							s->Connect ();
-						}
-						else
-						{
-							auto onEstablished = s->GetOnEstablished ();
-							if (onEstablished)
-								s->SetOnEstablished ([session, s, relayTag, onEstablished]()
-									{
-										onEstablished ();
-										s->Introduce (session, relayTag);
-									});
-							else
-								s->SetOnEstablished ([session, s, relayTag]() {s->Introduce (session, relayTag); });
-						}
-					}
+					s = std::make_shared<SSU2Session> (*this, r, addr);
+					s->SetOnEstablished ([session, s, relayTag]() { s->Introduce (session, relayTag); });
+					s->Connect ();
+				}
+				else
+				{
+					auto onEstablished = s->GetOnEstablished ();
+					if (onEstablished)
+						s->SetOnEstablished ([session, s, relayTag, onEstablished]()
+							{
+								onEstablished ();
+								s->Introduce (session, relayTag);
+							});
+					else
+						s->SetOnEstablished ([session, s, relayTag]() {s->Introduce (session, relayTag); });
 				}
 			}
+			else
+				session->Done ();
 		}
 		else
 		{
 			// introducers not found, try to request them
-			for (auto& it: address->ssu->introducers)
-				if (it.iTag && ts < it.iExp)
-					i2p::data::netdb.RequestDestination (it.iH);
+			for (auto& it: newRouters)
+				i2p::data::netdb.RequestDestination (it);
+			session->Done (); // don't wait for connect timeout
 		}
 	}
 
@@ -793,14 +998,16 @@ namespace transport
 		if (!router) return false;
 		auto addr = v4 ? router->GetSSU2V4Address () : router->GetSSU2V6Address ();
 		if (!addr) return false;
-		auto it = m_SessionsByRouterHash.find (router->GetIdentHash ());
-		if (it != m_SessionsByRouterHash.end ())
+		auto session = FindSession (router->GetIdentHash ());
+		if (session)
 		{
-			auto s = it->second;
-			if (it->second->IsEstablished ())
-				GetService ().post ([s]() { s->SendPeerTest (); });
+			auto remoteAddr = session->GetAddress ();
+			if (!remoteAddr || !remoteAddr->IsPeerTesting () ||
+			    (v4 && !remoteAddr->IsV4 ()) || (!v4 && !remoteAddr->IsV6 ())) return false;   
+			if (session->IsEstablished ())
+				GetService ().post ([session]() { session->SendPeerTest (); });
 			else
-				s->SetOnEstablished ([s]() { s->SendPeerTest (); });
+				session->SetOnEstablished ([session]() { session->SendPeerTest (); });
 			return true;
 		}
 		else
@@ -810,7 +1017,8 @@ namespace transport
 
 	void SSU2Server::ScheduleTermination ()
 	{
-		m_TerminationTimer.expires_from_now (boost::posix_time::seconds(SSU2_TERMINATION_CHECK_TIMEOUT));
+		m_TerminationTimer.expires_from_now (boost::posix_time::seconds(
+			SSU2_TERMINATION_CHECK_TIMEOUT + m_Rng () % SSU2_TERMINATION_CHECK_TIMEOUT_VARIANCE));
 		m_TerminationTimer.async_wait (std::bind (&SSU2Server::HandleTerminationTimer,
 			this, std::placeholders::_1));
 	}
@@ -820,17 +1028,20 @@ namespace transport
 		if (ecode != boost::asio::error::operation_aborted)
 		{
 			auto ts = i2p::util::GetSecondsSinceEpoch ();
-			for (auto it = m_PendingOutgoingSessions.begin (); it != m_PendingOutgoingSessions.end ();)
+
 			{
-				if (it->second->IsTerminationTimeoutExpired (ts))
+				std::lock_guard<std::mutex> l(m_PendingOutgoingSessionsMutex);
+				for (auto it = m_PendingOutgoingSessions.begin (); it != m_PendingOutgoingSessions.end ();)
 				{
-					//it->second->Terminate ();
-					std::unique_lock<std::mutex> l(m_PendingOutgoingSessionsMutex);
-					it = m_PendingOutgoingSessions.erase (it);
+					if (it->second->IsTerminationTimeoutExpired (ts))
+					{
+						//it->second->Terminate ();
+						it = m_PendingOutgoingSessions.erase (it);
+					}
+					else
+						it++;
 				}
-				else
-					it++;
-			}
+			}	
 
 			for (auto it: m_Sessions)
 			{
@@ -848,14 +1059,6 @@ namespace transport
 					it.second->CleanUp (ts);
 			}
 
-			for (auto it = m_SessionsByRouterHash.begin (); it != m_SessionsByRouterHash.begin ();)
-			{
-				if (it->second && it->second->GetState () == eSSU2SessionStateTerminated)
-					it = m_SessionsByRouterHash.erase (it);
-				else
-					it++;
-			}
-
 			ScheduleTermination ();
 		}
 	}
@@ -874,12 +1077,23 @@ namespace transport
 			auto ts = i2p::util::GetSecondsSinceEpoch ();
 			for (auto it = m_Relays.begin (); it != m_Relays.begin ();)
 			{
-				if (it->second && it->second->GetState () == eSSU2SessionStateTerminated)
+				if (it->second.expired ())
 					it = m_Relays.erase (it);
 				else
 					it++;
 			}
 
+			for (auto it = m_PeerTests.begin (); it != m_PeerTests.end ();)
+			{
+				if (ts > it->second.second + SSU2_PEER_TEST_EXPIRATION_TIMEOUT || it->second.first.expired ())
+				{
+					LogPrint (eLogInfo, "SSU2: Peer test nonce ", it->first, " was not responded in ", SSU2_PEER_TEST_EXPIRATION_TIMEOUT, " seconds or session invalid. Deleted");
+					it = m_PeerTests.erase (it);
+				}
+				else
+					it++;
+			}
+			
 			for (auto it = m_IncomingTokens.begin (); it != m_IncomingTokens.end (); )
 			{
 				if (ts > it->second.second)
@@ -896,7 +1110,35 @@ namespace transport
 					it++;
 			}
 
+			for (auto it = m_ConnectedRecently.begin (); it != m_ConnectedRecently.end (); )
+			{
+				if (ts > it->second + SSU2_HOLE_PUNCH_EXPIRATION)
+					it = m_ConnectedRecently.erase (it);
+				else
+					it++;
+			}	
+
+			for (auto it = m_RequestedPeerTests.begin (); it != m_RequestedPeerTests.end ();)
+			{
+				if (ts > it->second.second + SSU2_PEER_TEST_EXPIRATION_TIMEOUT)
+					it = m_RequestedPeerTests.erase (it);
+				else
+					it++;
+			}
+
+			{
+				std::lock_guard<std::mutex> l(m_SessionsByRouterHashMutex);
+				for (auto it = m_SessionsByRouterHash.begin (); it != m_SessionsByRouterHash.begin ();)
+				{
+					if (it->second.expired ())
+						it = m_SessionsByRouterHash.erase (it);
+					else
+						it++;
+				}
+			}	
+			
 			m_PacketsPool.CleanUpMt ();
+			m_PacketsArrayPool.CleanUpMt ();
 			m_SentPacketsPool.CleanUp ();
 			m_IncompleteMessagesPool.CleanUp ();
 			m_FragmentsPool.CleanUp ();
@@ -906,8 +1148,9 @@ namespace transport
 
 	void SSU2Server::ScheduleResend (bool more)
 	{
-		m_ResendTimer.expires_from_now (boost::posix_time::milliseconds (more ? SSU2_RESEND_CHECK_MORE_TIMEOUT :
-			(SSU2_RESEND_CHECK_TIMEOUT + rand () % SSU2_RESEND_CHECK_TIMEOUT_VARIANCE)));
+		m_ResendTimer.expires_from_now (boost::posix_time::milliseconds (more ? 
+		    (SSU2_RESEND_CHECK_MORE_TIMEOUT + m_Rng () % SSU2_RESEND_CHECK_MORE_TIMEOUT_VARIANCE):
+			(SSU2_RESEND_CHECK_TIMEOUT + m_Rng () % SSU2_RESEND_CHECK_TIMEOUT_VARIANCE)));
 		m_ResendTimer.async_wait (std::bind (&SSU2Server::HandleResendTimer,
 			this, std::placeholders::_1));
 	}
@@ -920,7 +1163,8 @@ namespace transport
 			auto ts = i2p::util::GetMillisecondsSinceEpoch ();
 			for (auto it: m_Sessions)
 			{
-				resentPacketsNum += it.second->Resend (ts);
+				if (ts >= it.second->GetLastResendTime () + SSU2_RESEND_CHECK_TIMEOUT)
+					resentPacketsNum += it.second->Resend (ts);
 				if (resentPacketsNum > SSU2_MAX_RESEND_PACKETS) break;
 			}
 			for (auto it: m_PendingOutgoingSessions)
@@ -977,97 +1221,98 @@ namespace transport
 		return ret;
 	}
 
-	std::list<std::shared_ptr<SSU2Session> > SSU2Server::FindIntroducers (int maxNumIntroducers,
-		bool v4, const std::set<i2p::data::IdentHash>& excluded) const
+	std::vector<std::shared_ptr<SSU2Session> > SSU2Server::FindIntroducers (int maxNumIntroducers,
+		bool v4, const std::unordered_set<i2p::data::IdentHash>& excluded)
 	{
-		std::list<std::shared_ptr<SSU2Session> > ret;
+		std::vector<std::shared_ptr<SSU2Session> > ret;
+		if (maxNumIntroducers <= 0 || m_Sessions.empty ()) return ret;
+		
+		std::vector<std::shared_ptr<SSU2Session> > eligible;
+		eligible.reserve (m_Sessions.size ()/2);
+		auto ts = i2p::util::GetSecondsSinceEpoch ();
 		for (const auto& s : m_Sessions)
 		{
 			if (s.second->IsEstablished () && (s.second->GetRelayTag () && s.second->IsOutgoing ()) &&
+			    ts < s.second->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_DURATION/2 &&
 			    !excluded.count (s.second->GetRemoteIdentity ()->GetIdentHash ()) &&
 			    ((v4 && (s.second->GetRemoteTransports () & i2p::data::RouterInfo::eSSU2V4)) ||
 			    (!v4 && (s.second->GetRemoteTransports () & i2p::data::RouterInfo::eSSU2V6))))
-				ret.push_back (s.second);
-		}
-		if ((int)ret.size () > maxNumIntroducers)
-		{
-			// shink ret randomly
-			int sz = ret.size () - maxNumIntroducers;
-			for (int i = 0; i < sz; i++)
-			{
-				auto ind = rand () % ret.size ();
-				auto it = ret.begin ();
-				std::advance (it, ind);
-				ret.erase (it);
-			}
+				eligible.push_back (s.second);
 		}
+		
+		std::sample (eligible.begin(), eligible.end(), std::back_inserter(ret), maxNumIntroducers, m_Rng);
 		return ret;
 	}
 
 	void SSU2Server::UpdateIntroducers (bool v4)
 	{
 		uint32_t ts = i2p::util::GetSecondsSinceEpoch ();
-		std::list<i2p::data::IdentHash> newList, impliedList;
+		std::list<std::pair<i2p::data::IdentHash, uint32_t> > newList, impliedList;
 		auto& introducers = v4 ? m_Introducers : m_IntroducersV6;
-		std::set<i2p::data::IdentHash> excluded;
-		for (const auto& it : introducers)
+		std::unordered_set<i2p::data::IdentHash> excluded;
+		for (const auto& [ident, tag] : introducers)
 		{
-			std::shared_ptr<SSU2Session> session;
-			auto it1 = m_SessionsByRouterHash.find (it);
-			if (it1 != m_SessionsByRouterHash.end ())
-			{
-				session = it1->second;
-				excluded.insert (it);
-			}
-			if (session && session->IsEstablished () && session->GetRelayTag () && session->IsOutgoing ()) // still session with introducer?
-			{
-				if (ts < session->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_EXPIRATION)
+			std::shared_ptr<SSU2Session> session = FindSession (ident);
+			if (session)
+				excluded.insert (ident);
+			if (session)
+			{	
+				if (session->IsEstablished () && session->GetRelayTag () && session->IsOutgoing () && // still session with introducer?
+					ts < session->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_EXPIRATION)
 				{	
 					session->SendKeepAlive ();
 					if (ts < session->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_DURATION)	
-						newList.push_back (it);
+					{	
+						newList.push_back ({ident, session->GetRelayTag ()});
+						if (tag != session->GetRelayTag ())
+						{
+							LogPrint (eLogDebug, "SSU2: Introducer session to  ", session->GetIdentHashBase64() , " was replaced. iTag ", tag, "->", session->GetRelayTag ());
+							i2p::context.UpdateSSU2Introducer (ident, v4, session->GetRelayTag (),
+								session->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_EXPIRATION);
+						}	
+					}
 					else	
 					{	
-						impliedList.push_back (it); // keep in introducers list, but not publish
+						impliedList.push_back ({ident, session->GetRelayTag ()}); // keep in introducers list, but not publish
 						session = nullptr;	
-					}		
+					}
 				}	
 				else
 					session = nullptr;
-			}
+			}	
+			
 			if (!session)
-				i2p::context.RemoveSSU2Introducer (it, v4);
+				i2p::context.RemoveSSU2Introducer (ident, v4);
 		}
+		int numOldSessions = 0;
 		if (newList.size () < SSU2_MAX_NUM_INTRODUCERS)
 		{
 			auto sessions = FindIntroducers (SSU2_MAX_NUM_INTRODUCERS - newList.size (), v4, excluded);
-			if (sessions.empty () && !introducers.empty ())
+			if (sessions.empty () && !impliedList.empty ())
 			{
-				// bump creation time for previous introducers if no new sessions found
 				LogPrint (eLogDebug, "SSU2: No new introducers found. Trying to reuse existing");
-				impliedList.clear ();
-				for (auto& it : introducers)
+				for (const auto& it : impliedList)
 				{
-					auto it1 = m_SessionsByRouterHash.find (it);
-					if (it1 != m_SessionsByRouterHash.end ())
+					auto session = FindSession (it.first);
+					if (session)
 					{
-						auto session = it1->second;
-						if (session->IsEstablished () && session->GetRelayTag () && session->IsOutgoing ())
-						{
-							session->SetCreationTime (session->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_DURATION);
-							if (std::find (newList.begin (), newList.end (), it) == newList.end ())
-								sessions.push_back (session);
-						}
+						if (std::find_if (newList.begin (), newList.end (), 
+						    [&ident = it.first](const auto& s){ return ident == s.first; }) == newList.end ())
+						{	
+							sessions.push_back (session);
+							numOldSessions++;
+						}	
 					}
 				}
+				impliedList.clear ();
 			}
 
 			for (const auto& it : sessions)
 			{
-				uint32_t tag = it->GetRelayTag ();
+				uint32_t tag = it->GetRelayTag ();		
 				uint32_t exp = it->GetCreationTime () + SSU2_TO_INTRODUCER_SESSION_EXPIRATION;
-				if (!tag || ts + SSU2_TO_INTRODUCER_SESSION_DURATION/2 > exp)
-					continue; // don't pick too old session for introducer	
+				if (!tag && ts >= exp) 
+					continue; // don't publish expired introducer
 				i2p::data::RouterInfo::Introducer introducer;
 				introducer.iTag = tag;
 				introducer.iH = it->GetRemoteIdentity ()->GetIdentHash ();
@@ -1077,16 +1322,28 @@ namespace transport
 				{
 					LogPrint (eLogDebug, "SSU2: Introducer added ", it->GetRelayTag (), " at ",
 						i2p::data::GetIdentHashAbbreviation (it->GetRemoteIdentity ()->GetIdentHash ()));
-					newList.push_back (it->GetRemoteIdentity ()->GetIdentHash ());
+					newList.push_back ({ it->GetRemoteIdentity ()->GetIdentHash (), tag });
+					it->SendKeepAlive ();
 					if (newList.size () >= SSU2_MAX_NUM_INTRODUCERS) break;
 				}
 			}
 		}
 		introducers = newList;
 
-		if (introducers.size () < SSU2_MAX_NUM_INTRODUCERS)
+		if (introducers.size () < SSU2_MAX_NUM_INTRODUCERS || numOldSessions)
 		{
-			for (auto i = introducers.size (); i < SSU2_MAX_NUM_INTRODUCERS; i++)
+			// we need to create more sessions with relay tag
+			
+			// exclude all existing sessions
+			excluded.clear ();
+			{
+				std::lock_guard<std::mutex> l(m_SessionsByRouterHashMutex);
+				for (const auto& [ident, s] : m_SessionsByRouterHash)
+					excluded.insert (ident);
+			}	
+
+			// sesssion about to expire are not counted
+			for (auto i = introducers.size (); i < SSU2_MAX_NUM_INTRODUCERS + numOldSessions; i++)
 			{
 				auto introducer = i2p::data::netdb.GetRandomSSU2Introducer (v4, excluded);
 				if (introducer)
@@ -1113,7 +1370,7 @@ namespace transport
 		if (m_IsPublished)
 		{
 			m_IntroducersUpdateTimer.expires_from_now (boost::posix_time::seconds(
-				SSU2_KEEP_ALIVE_INTERVAL + rand () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE));
+				SSU2_KEEP_ALIVE_INTERVAL + m_Rng () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE));
 			m_IntroducersUpdateTimer.async_wait (std::bind (&SSU2Server::HandleIntroducersUpdateTimer,
 				this, std::placeholders::_1, true));
 		}
@@ -1127,7 +1384,7 @@ namespace transport
 			i2p::context.ClearSSU2Introducers (true);
 			m_Introducers.clear ();
 			m_IntroducersUpdateTimer.expires_from_now (boost::posix_time::seconds(
-				(SSU2_KEEP_ALIVE_INTERVAL + rand () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE)/2));
+				(SSU2_KEEP_ALIVE_INTERVAL + m_Rng () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE)/2));
 			m_IntroducersUpdateTimer.async_wait (std::bind (&SSU2Server::HandleIntroducersUpdateTimer,
 				this, std::placeholders::_1, true));
 		}
@@ -1138,7 +1395,7 @@ namespace transport
 		if (m_IsPublished)
 		{
 			m_IntroducersUpdateTimerV6.expires_from_now (boost::posix_time::seconds(
-				SSU2_KEEP_ALIVE_INTERVAL + rand () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE));
+				SSU2_KEEP_ALIVE_INTERVAL + m_Rng () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE));
 			m_IntroducersUpdateTimerV6.async_wait (std::bind (&SSU2Server::HandleIntroducersUpdateTimer,
 				this, std::placeholders::_1, false));
 		}
@@ -1152,7 +1409,7 @@ namespace transport
 			i2p::context.ClearSSU2Introducers (false);
 			m_IntroducersV6.clear ();
 			m_IntroducersUpdateTimerV6.expires_from_now (boost::posix_time::seconds(
-				(SSU2_KEEP_ALIVE_INTERVAL + rand () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE)/2));
+				(SSU2_KEEP_ALIVE_INTERVAL + m_Rng () % SSU2_KEEP_ALIVE_INTERVAL_VARIANCE)/2));
 			m_IntroducersUpdateTimerV6.async_wait (std::bind (&SSU2Server::HandleIntroducersUpdateTimer,
 				this, std::placeholders::_1, false));
 		}
